//Michael Good, 12/7/2011
//Veronica Hull, 12/7/2011
//Renee Sundrud, CPS 162
//
// VERSION: M5f:V3f:S0

#include "BoardPosition.h"

//Holds a board position of two sides of pieces

using namespace std;

BoardPosition::checker mine[12];
BoardPosition::checker AI[12];

//CONSTRUCTOR
// -----------------------------------------------------------------------------
//constructor; initializes all pieces to beginning of game
BoardPosition::BoardPosition() //creates a new game
{
    int odd = 0;
    int ctr = 0;
    
    for (int yPos = 7; yPos > 4; yPos--)
    {
        for (int xPos = 0; xPos < 8; xPos += 2)
        {
            mine[ctr].king = false;
            mine[ctr].exists = true;
            mine[ctr].x = xPos + odd;
            mine[ctr].y = yPos;
            mine[ctr].side = -1;
            ctr++;
        }
        if (odd)
            odd = 0;
        else
            odd = 1;
    }
    
    ctr = 0;
    for (int yPos = 2; yPos > -1; yPos--)
    {
        for (int xPos = 0; xPos < 8; xPos += 2)
        {
            AI[ctr].king = false;
            AI[ctr].exists = true;
            AI[ctr].x = xPos + odd;
            AI[ctr].y = yPos;
            AI[ctr].side = 1;
            ctr++;
        }
        if (odd)
            odd = 0;
        else
            odd = 1;
    }
}

//AI FUNCTIONS
// -----------------------------------------------------------------------------
//finds the best move for the computer to go (higher levels give better moves)
BoardPosition BoardPosition::bestMove (int level, bool player, bool first)
{
    double bestCandidate = -9999; //points of best move
    double curCandidate = 0;      //points of current move
    int who;                      //which side we are running
    int random;                   //randomize between best moves
    
    vector<BoardPosition> moves;
    vector<BoardPosition> possibilities;
    vector<BoardPosition> candidates;
    vector<int> references;
    
    if (player)
    {
        moves = stackMoves("me");
        who = -1;
    }
    else
    {
        moves = stackMoves("ai");
        who = 1;
        level -= 1;
    }
    
    //if there are no moves available
    if (moves.size() == 0)
       return *this;
    
    //check if we are at the end of the recursion loop (level=0)
    if (level != 0)
    {
       //to find the best move, the computer must anticipate the opponent's best
       //move... play the other side in the next recursive function
       if (player)
          player = false;
       else
           player = true;
           
       //make a list of all the best possibilities from inner recursions
       for (int ctr = 0; ctr < moves.size(); ctr++)
       {
           //testOutput();
           possibilities.push_back(moves[ctr].bestMove(level, player, false));
           //system("PAUSE");
       }
       
       //find the best move from the best possibilities
       for (int ctr = 0; ctr < possibilities.size(); ctr++)
       {
           curCandidate = possibilities[ctr].pointValue(who);
           if (curCandidate > bestCandidate)
           {
              candidates.clear();
              references.clear();
              bestCandidate = curCandidate;
              candidates.push_back(possibilities[ctr]);
              references.push_back(ctr);
           }
           else if (curCandidate == bestCandidate)
           {
                candidates.push_back(possibilities[ctr]);
                references.push_back(ctr);
           }
       }
       
       //if this is not the first loop, return the best board position
       if (!first)
       {
          srand(time(NULL));
          random = rand() % candidates.size();
          return candidates[random];
       }
       else //else, return the board position with the computer's move
       {
           srand(time(NULL));
           random = rand() % references.size();
           return moves[references[random]];
       }
    }
    else
    {
        srand(time(NULL));
        random = rand() % moves.size();
        return moves[random];
    }
}

// -----------------------------------------------------------------------------
//returns the points value of a position
double BoardPosition::pointValue(int side)
{
       double returnVal;
       
       for (int ctr = 0; ctr < 12; ctr++)
       {
           if (mine[ctr].exists)
           {
              returnVal -= side;
              if (mine[ctr].king)
                 returnVal -= side;
           }
           if (AI[ctr].exists)
           {
              returnVal += side;
              if (AI[ctr].king)
                 returnVal += side;
           }
       }
       
       return returnVal;
}

//OPERATOR FUNCTIONS
// -----------------------------------------------------------------------------
//compares two boards to see if they have the same position
bool BoardPosition::operator==(BoardPosition toCheck)
{
     bool match = true;
     checker *checkSpot;
     
     if (pieceCount("me") == toCheck.pieceCount("me")
        && pieceCount("AI") == toCheck.pieceCount("AI"))
     {
        for (int ctr = 0; ctr < 12; ctr++)
        {
            if (mine[ctr].exists)
            {
               checkSpot = toCheck.getChecker(mine[ctr].x, mine[ctr].y);
               if (checkSpot)
               {
                  if (checkSpot->side != mine[ctr].side)
                     match = false;
               }
               else
               {
                   match = false;
               }
            }
            if (AI[ctr].exists)
            {
               checkSpot = toCheck.getChecker(AI[ctr].x, AI[ctr].y);
               if (checkSpot)
               {
                  if (checkSpot->side != AI[ctr].side)
                     match = false;
               }
               else
               {
                   match = false;
               }
            }
        }
     }
     else
     {
         match = false;
     }
         
     return match;
}

// -----------------------------------------------------------------------------
//compares just one side of checkers on two boards
bool BoardPosition::operator== (checker compare[12])
{
     int side = compare[0].side;
     bool areEqual = false;
     BoardPosition checkPosition;
     
     for (int ctr = 0; ctr < 12; ctr++)
     {
         if (side == mine[0].side)
         {
            checkPosition.mine[ctr] = compare[ctr];
            checkPosition.AI[ctr] = AI[ctr];
         }
         if (side == AI[0].side)
         {
            checkPosition.mine[ctr] = AI[ctr];
            checkPosition.AI[ctr] = compare[ctr];
         }
     }
     
     if (checkPosition == *this)
        areEqual = true;
        
     return areEqual;
}

// -----------------------------------------------------------------------------
//sets one board to be a copy of another board
void BoardPosition::operator= (BoardPosition *toSet)
{
     for (int ctr = 0; ctr < 12; ctr++)
     {
         mine[ctr].king = toSet->mine[ctr].king;
         mine[ctr].exists = toSet->mine[ctr].exists;
         mine[ctr].x = toSet->mine[ctr].x;
         mine[ctr].y = toSet->mine[ctr].y;
         mine[ctr].side = toSet->mine[ctr].side;
         
         AI[ctr].king = toSet->AI[ctr].king;
         AI[ctr].exists = toSet->AI[ctr].exists;
         AI[ctr].x = toSet->AI[ctr].x;
         AI[ctr].y = toSet->AI[ctr].y;
         AI[ctr].side = toSet->AI[ctr].side;
     }
}

//OTHER FUNCTIONS
// -----------------------------------------------------------------------------
//stacks all the moves for each checker on a given side
vector<BoardPosition> BoardPosition::stackMoves(string who)
{
     bool jumps = false;
     int newJumps = false;
     vector<BoardPosition> moves;
     vector<BoardPosition> newMoves;
     
     for (int ctr = 0; ctr < 12; ctr++)
     {   
         if (who == "me")
         {
            if (mine[ctr].exists)
            {
               newMoves = findMoves(&mine[ctr], &newJumps); //get new moves
               
               if (newJumps && !jumps) //first jump encountered
               {
                  jumps = true;
                  moves.clear();
               }
               else if (!newJumps && jumps) //more jumps encountered
               {
                    newMoves.clear();
               }
            
               while (!newMoves.empty())
               {
                     moves.push_back(newMoves.back());
                     newMoves.pop_back();
               }
            }
         }
         else if (who == "ai")
         {
            if (AI[ctr].exists)
            {
               newMoves = findMoves(&AI[ctr], &newJumps); //get new moves
               
               if (newJumps && !jumps) //first jump encountered
               {
                  jumps = true;
                  moves.clear();
               }
               else if (!newJumps && jumps) //more jumps encountered
               {
                    newMoves.clear();
               }
            
               while (!newMoves.empty())
               {
                     moves.push_back(newMoves.back());
                     newMoves.pop_back();
               }
            }
         }
     }
     
     return moves;
}

// -----------------------------------------------------------------------------
//returns a vector of all valid moves for a given checker
vector<BoardPosition> BoardPosition::findMoves(checker *piece, int *jumps)
{
     int occupiedBy;
     bool checkJump;
     vector<BoardPosition> moves;
     checker *toCheck;
     BoardPosition newPosition;
     
         //check jumps
         checkJump = isJump(piece);
         if (checkJump)
         {
            if (!*jumps)
            {
               *jumps = true;
               moves.clear();
            }
            
            newPosition = this;
            newPosition.getJumps(piece, moves);
         }
     
         //if there are no jumps
         if (!*jumps)
         {
            //check left
            occupiedBy = occupied(piece->x - 1, piece->y + piece->side);
            if (!occupiedBy)
            {
               newPosition = this;
               toCheck = newPosition.getChecker(piece->x, piece->y);
               toCheck->x -= 1;
               toCheck->y += piece->side;
               if (toCheck->y == (3.5 + piece->side*3.5))
                  toCheck->king = true;
               //assign newPosition to change checkers
               moves.push_back(newPosition);
            }
     
            //check right
            occupiedBy = occupied(piece->x + 1, piece->y + piece->side);
            if (!occupiedBy)
            {
               newPosition = this;
               toCheck = newPosition.getChecker(piece->x, piece->y);
               toCheck->x += 1;
               toCheck->y += piece->side;
               if (toCheck->y == (3.5 + piece->side*3.5))
                  toCheck->king = true;
               //assign newPosition to change checkers
               moves.push_back(newPosition);
            }
            
            //if the piece is a king, then check backwards too
            if (piece->king)
            {
               //check left
               occupiedBy = occupied(piece->x - 1, piece->y - piece->side);
               if (!occupiedBy)
               {
                  newPosition = this;
                  toCheck = newPosition.getChecker(piece->x, piece->y);
                  toCheck->x -= 1;
                  toCheck->y -= piece->side;
                  //assign newPosition to change checkers
                  moves.push_back(newPosition);
               }
     
               //check right
               occupiedBy = occupied(piece->x + 1, piece->y - piece->side);
               if (!occupiedBy)
               {
                  newPosition = this;
                  toCheck = newPosition.getChecker(piece->x, piece->y);
                  toCheck->x += 1;
                  toCheck->y -= piece->side;
                  //assign newPosition to change checkers
                  moves.push_back(newPosition);
               }
            }
         }
         
         return moves;             
}

// -----------------------------------------------------------------------------
//checks if a given square is occupied by a certain checker
// occupied = -1... player's checker
// occupied = 1... cpu's checker
// occupied = 2... endge of board
int BoardPosition::occupied (int x, int y)
{
     int occupiedBy = 0;
     
     //test for edge of screen
     if (x<0 || x>7 || y<0 || y>7)
     {
        occupiedBy = 2;
     }

     //test for other checkers
     if (!occupiedBy)
     {
        for (int ctr = 0; ctr < 12; ctr++)
        {
            if ((mine[ctr].x == x) && (mine[ctr].y == y) && mine[ctr].exists)
            {
               occupiedBy = -1;
               ctr = 13;
            }
        }
        if (occupiedBy == 0)
        {
           for (int ctr = 0; ctr < 12; ctr++)
           {
               if (AI[ctr].x == x && AI[ctr].y == y && AI[ctr].exists)
               {
                  occupiedBy = 1;
                  ctr = 13;
               }
           }
        }
     }
     
     return occupiedBy;
}

// -----------------------------------------------------------------------------
//returns a pointer to a checker
BoardPosition::checker *BoardPosition::getChecker (int x, int y)
{
        checker *toReturn = NULL;
        
        for (int ctr = 0; ctr < 12; ctr++)
        {
            if (mine[ctr].x == x && mine[ctr].y == y && mine[ctr].exists)
            {
               toReturn = &mine[ctr];
            }
            else if (AI[ctr].x == x && AI[ctr].y == y && AI[ctr].exists)
            {
                 toReturn = &AI[ctr];
            }
        }
        
        return toReturn;
}

// -----------------------------------------------------------------------------
//returns how many pieces a side has
int BoardPosition::pieceCount(string who)
{
    int count = 0;
    for (int ctr = 0; ctr < 12; ctr++)
    {
        if (who == "me")
        {
           if (mine[ctr].exists)
              count++;
        }
        else if (who == "ai")
        {
             if (AI[ctr].exists)
                count++;
        }
    }
    return count;
}

//PRIVATE FUNCTIONS
// -----------------------------------------------------------------------------
//tests if there are any jumps or not
bool BoardPosition::isJump(checker *piece)
{
     int x = piece->x;
     int y = piece->y;
     int side = piece->side;
     bool jump = false;
     checker *toCheck;
     
     //can we jump forwards/left?
     toCheck = getChecker(x-1, y+side);
     if (toCheck)
        if ((toCheck->side != side) && !occupied(x-2, y+(side*2)))
        {
           jump = true;
        }
           
     //can we jump forwards/right?
     toCheck = getChecker(x+1, y+side);
     if (toCheck)
        if ((toCheck->side != side) && !occupied(x+2, y+(side*2)))
        {
           jump = true;
        }
           
     if (piece->king)
     {
        //can we jump backwards/left?
        toCheck = getChecker(x-1, y-side);
        if (toCheck)
           if ((toCheck->side != side) && !occupied(x-2, y-(side*2)))
              jump = true;
           
        //can we jump backwards/right?
        toCheck = getChecker(x+1, y-side);
        if (toCheck)
           if ((toCheck->side != side) && !occupied(x+2, y-(side*2)))
              jump = true;
     }
     
     return jump;
}

// -----------------------------------------------------------------------------
//0 = no places to jump
bool BoardPosition::getJumps (checker *piece, vector<BoardPosition> &moves)
{    
     checker *checkPiece; //piece that will be jumped
     checker *jumpPiece;  //piece that will do the jumping
     BoardPosition newPosition;
     int checkPlace1;
     int checkPlace2;
     bool toReturn = false;
     int returnCheck;
     
     //check for jumps to the left
     checkPlace1 = occupied(piece->x-1, piece->y + piece->side);
     checkPlace2 = occupied(piece->x-2, piece->y + (piece->side*2));

     if (checkPlace1 && !checkPlace2 && checkPlace1 != piece->side)
     {
        newPosition = this;
        toReturn = true;
            
        checkPiece = newPosition.getChecker(piece->x-1, piece->y + piece->side);
        jumpPiece = newPosition.getChecker(piece->x, piece->y);
            
        //delete checker from board position
        checkPiece->exists = false;
        
        //move checker
        jumpPiece->x -= 2;
        jumpPiece->y += jumpPiece->side*2;
            
        //check for more jumps at this position
        if (newPosition.getJumps(jumpPiece, moves))
        {
           toReturn = true;
        }
        else //queue jump if end of jump chain
        {
            if (jumpPiece->y == (3.5 + piece->side*3.5))
                  jumpPiece->king = true; 
             
            moves.push_back(newPosition);
        }
     }
     
     //check for jumps to the right
     checkPlace1 = occupied(piece->x+1, piece->y + piece->side);
     checkPlace2 = occupied(piece->x+2, piece->y + (piece->side*2));
     
     if (checkPlace1 && !checkPlace2 && checkPlace1 != piece->side)
     {
        newPosition = this;
        toReturn = true;
            
        checkPiece = newPosition.getChecker(piece->x+1, piece->y + piece->side);
        jumpPiece = newPosition.getChecker(piece->x, piece->y);
        
        //delete checker from board position
        checkPiece->exists = false;
        
        //move checker
        jumpPiece->x += 2;
        jumpPiece->y += jumpPiece->side*2;
        
        //check for more jumps at this position
        if (newPosition.getJumps(jumpPiece, moves))
        {
           toReturn = true;
        }
        else //queue jump if end of jump chain
        {
            if (jumpPiece->y == (3.5 + piece->side*3.5))
                  jumpPiece->king = true;
             
            moves.push_back(newPosition);
        }
     }
     
     //if the piece is a king, check for jumps going backwards too
     if (piece->king)
     {
        //check for jumps to the left
        checkPlace1 = occupied(piece->x-1, piece->y - piece->side);
        checkPlace2 = occupied(piece->x-2, piece->y - (piece->side*2));
         
        if (checkPlace1 && !checkPlace2 && checkPlace1 != piece->side)
        {
           newPosition = this;
           toReturn = true;
            
           checkPiece = newPosition.getChecker(piece->x-1,piece->y-piece->side);
           jumpPiece = newPosition.getChecker(piece->x, piece->y);
            
           //delete checker from board position
           checkPiece->exists = false;
           //move checker
           jumpPiece->x -= 2;
           jumpPiece->y -= jumpPiece->side*2;
            
           //check for more jumps at this position
           if (newPosition.getJumps(jumpPiece, moves))
           {
              toReturn = true;
           }
           else //queue jump if end of jump chain
           {
               moves.push_back(newPosition);
           }
        }
     
        //check for jumps to the right
        checkPlace1 = occupied(piece->x+1, piece->y - piece->side);
        checkPlace2 = occupied(piece->x+2, piece->y -(piece->side*2));
     
        if (checkPlace1 && !checkPlace2 && checkPlace1 != piece->side)
        {
           newPosition = this;
           toReturn = true;
            
           checkPiece = newPosition.getChecker(piece->x+1,piece->y-piece->side);
           jumpPiece = newPosition.getChecker(piece->x, piece->y);
            
           //delete checker from board position
           checkPiece->exists = false;
           //move checker
           jumpPiece->x += 2;
           jumpPiece->y -= jumpPiece->side*2;
            
           //check for more jumps at this position
           if (newPosition.getJumps(jumpPiece, moves))
           {
              toReturn = true;
           }
           else //queue jump if end of jump chain
           {      
               moves.push_back(newPosition);
           }
        }
     }
     
     return toReturn;
}
