//Michael Good, 12/7/2011
//Veronica Hull, 12/7/2011
//Renee Sundrud, CPS 162
//
// VERSION: M5f:V3f:S0
//
//
// ****************
// **  CHECKERS  **
// ****************
//
//       Checkers is a fun little game where the computer will beat, er, I mean
// play, a human in a game of checkers. Difficulty can be scaled to the skill
// level of the player.
//
// CREDITS:
//
//   Michael Good
//     - Move Validation, Artificial Intelligence, Various Functions
//
//   Veronica Hull 
//     - Input/Output, Program Cleanup, Testing

#include <iostream>
#include <iomanip>
#include "BoardPosition.h"

using namespace std;

const int MAX = 8;

//gets input, and returns whether the player wishes to quit
bool getInput(int &x1, int &x2, int &y1, int &y2);
//displays output
void showBoard(int[][MAX], BoardPosition &);
//finds which piece to display
char showPiece(int);

// -----------------------------------------------------------------------------
int main()
{
    //GAME SETUP
    int level;
    bool first;
    char input;
    int xBegin;
    int yBegin;
    int xEnd;
    int yEnd;
    bool gameOver = false;
    bool validMove;
       
    BoardPosition myPosition;
    BoardPosition checkPosition;
    BoardPosition::checker *piece;
    vector<BoardPosition> moves;
    
    vector<BoardPosition> getPositions;
    int board[MAX][MAX] = {{0,1,0,1,0,1,0,1},
                           {1,0,1,0,1,0,1,0},
                           {0,1,0,1,0,1,0,1},
                           {0,0,0,0,0,0,0,0},
                           {0,0,0,0,0,0,0,0},
                           {2,0,2,0,2,0,2,0},
                           {0,2,0,2,0,2,0,2},
                           {2,0,2,0,2,0,2,0}};
                           
    //get who is going first
    cout << "What level do you want to play at? ";
    cin >> level;
    cout << "Do you want the computer to go first? (Y/N) ";
    cin >> input;
    
    if (level < 1)
       level = 1;
    if (tolower(input) == 'y')
       first = false;
    else
        first = true;
                           
    //is the computer going first?
    if (!first)
    {
       myPosition = myPosition.bestMove(level);
       
       showBoard(board, myPosition);
    }
    
    //MAIN GAME LOOP
    while (!gameOver)
    {     
          //get (and validate) input
          validMove = false;
          while (!validMove)
          {
                showBoard(board, myPosition);
                gameOver = getInput (xBegin, yBegin, xEnd, yEnd); //get input
                if (gameOver)
                   validMove = true;
                
                checkPosition = myPosition;
          
                if (myPosition.occupied(xBegin, yBegin) == -1)
                {
                   piece = checkPosition.getChecker(xBegin, yBegin);
                   piece->x = xEnd;
                   piece->y = yEnd;
          
                   moves = myPosition.stackMoves("me");
          
                   for (int ctr = 0; ctr < moves.size(); ctr++)
                   {
                       if (checkPosition == moves[ctr].mine)
                       {
                          validMove = true;
                          //check if piece should be kinged
                          if (yEnd == 0)
                             piece->king = true;
                          
                          //update board
                          myPosition = moves[ctr];
                       }
                   }
                
                   if (!validMove)
                   {
                      cout << endl << "That is not a valid move!" << endl;
                      system("PAUSE");
                   }
                }
                else if (!gameOver)
                {
                    cout << endl << "That is not a valid move!" << endl;
                      system("PAUSE");
                }
                
                showBoard(board, myPosition);
                system("PAUSE");
          }
          
          //test if the game has been won
          if (myPosition.pieceCount("ai") * myPosition.pieceCount("me") == 0)
          {
             gameOver = true;
          }
          else //the computer makes their move...
          {
               //get computer's move
               myPosition = myPosition.bestMove(level);
               
               showBoard(board, myPosition);
               
               //test if the game is lost
               if (myPosition.pieceCount("ai") * myPosition.pieceCount("me")==0)
               {
                  gameOver = true;
               }
          }
    }
    
    //GAME OVER
    cout << endl;
    
    //if the computer has beat the player
    if (myPosition.pointValue(1) > myPosition.pointValue(-1))
         cout << "Haha, you lose!" << endl;
    //if the player has beat the computer
    else if (myPosition.pointValue(1) < myPosition.pointValue(-1))
         cout << "Nooo, you beat me!" << endl;
    //if there is a tie
    else
        cout << "Aw, I was getting ready to beat you, too!" << endl;
    
    
    system("PAUSE");
}

// -----------------------------------------------------------------------------
bool getInput(int &x1, int &y1, int &x2, int &y2)
{
     string input;
     cout << "Enter location of checker: ";
     cin >> input;
     
     if (input == "exit")
        return true;
     
     input += "!!";
     y1 = input.at(0);
     y1 -= 48;
     x1 = toupper(input.at(1));
     x1 -= 65;
     cout << "Enter where you want this checker to go: ";
     cin >> input;
     input += "!!";
     y2 = input.at(0);
     y2 -= 48;
     x2 = toupper(input.at(1));
     x2 -= 65;
     
     return false;
}

// -----------------------------------------------------------------------------
void showBoard(int matrix[MAX][MAX], BoardPosition &myPosition)
{
    char gamePiece;
    
    // reset matrix to zero
    for (int row = 0; row < 8; row++)
    {
        for (int col = 0; col < 8; col++)
        {
            matrix[row][col] = 0;    
        }
    }
    
    // assign player's checkers to board
    for (int index = 0; index < 12; index++)
    {
        int x;
        int y;
        
        y = myPosition.mine[index].x;
        x = myPosition.mine[index].y;
        
       if (myPosition.mine[index].exists)
       {
          if (!myPosition.mine[index].king)
             matrix[x][y] = 2;
          else
              matrix[x][y] = 4;                                    
       }
    }
    
    // assign computer's checkers to board
    for (int index = 0; index < 12; index++)
    {
        int x;
        int y;
        
        y = myPosition.AI[index].x;
        x = myPosition.AI[index].y;
        
       if (myPosition.AI[index].exists)
       {
          if (!myPosition.AI[index].king)
             matrix[x][y] = 1;
          else
              matrix[x][y] = 3;
       }                                    
    }
        
    system("CLS");
    
    cout << "INSTRUCTIONS:" << endl
<< "Enter the location of the piece you want to move, followed by the" << endl
<< "location that you want to move the piece to. Locations are" << endl
<< "referenced by their corresponding number, followed by their" << endl
<< "corresponding letter. (ex: 1A)" << endl << endl;
    
    cout << "|---|---|---|---|---|---|---|---|" << endl;
 
    for(int row = 0; row < 8; row++)
    {
        for(int column = 0; column < 8; column++)
        {
            gamePiece = showPiece(matrix[row][column]);
            if (column != 0)
               cout << "| " << gamePiece << " ";
            else
                cout << row << " " << gamePiece << " ";
        }
 
        cout << "|" << endl;
        if (row == 7)
           cout << "|-A-|-B-|-C-|-D-|-E-|-F-|-G-|-H-|" << endl;
        else
            cout << "|---|---|---|---|---|---|---|---|" << endl;
    }     
    
    cout << endl;
}

// -----------------------------------------------------------------------------
char showPiece(int num)
{
    switch(num)
    {
        case 0  :   return(' ');
                    break;
        case 1  :   return('w');
                    break;
        case 2  :   return('b');
                    break;
        case 3  :   return('W');
                    break;
        case 4  :   return('B');
                    break;
    }
}
